/*****************************************

ECWorkshop -- evolutionary computation
toolkit library

Web: http://qai.narod.ru/ecw/
Mailto: qai@mail.ru

Copyright (C) 2007 by Yury Tsoy


This program is free software; you can
redistribute it and/or modify it under the
terms of the GNU General Public License as
published by the Free Software Foundation;
either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that
it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU General Public License
for more details.

You should have received a copy of the GNU
General Public License along with this
program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330,
Boston, MA  02111-1307  USA

*/

#include "qevolutionaryalgorithm.h"

#include "../problems/include.h"

#include <algorithm>
#include <fstream>

using namespace ECWorkshop;

///////////////////////////////////////////////////////////
//
//	QEvolutionaryAlgorithm class
//
///////////////////////////////////////////////////////////

//---------------------------------------------------------
QEvolutionaryAlgorithm::QEvolutionaryAlgorithm (void)
: _popul(NULL), _selPopul(NULL), _children(NULL), _init(NULL),
_selection(NULL), _populationSizer(NULL), _xover(NULL), _mutation(NULL) {
	_factory = new QEAFactory (this);
}

//---------------------------------------------------------
QEvolutionaryAlgorithm::QEvolutionaryAlgorithm (const QEvolutionaryAlgorithm& argEA)
: _factory(argEA._factory), _params(argEA._params) {
	//if (argEA._popul != NULL) {
	//	_popul->assign (argEA._popul);
	//}
	//if (argEA._children != NULL) {
	//	_popul->assign (argEA._children);
	//}
	//if (argEA._selPopul != NULL) {
	//	_popul->assign (argEA._selPopul);
	//}
}

//---------------------------------------------------------
QEvolutionaryAlgorithm::~QEvolutionaryAlgorithm (void) {
	if (_factory != NULL) {
		delete _factory;
		_factory = NULL;
	}
	_popul = _selPopul = _children = NULL;
	_init = NULL;
	_selection = NULL;
	_xover = NULL;
	_mutation = NULL;

	if (_fout.is_open()) {
		_fout.close ();
	}
}

//---------------------------------------------------------
QEvolutionaryAlgorithm& QEvolutionaryAlgorithm::operator= (const QEvolutionaryAlgorithm& argEA) {
	_factory = argEA._factory;
	_params = argEA._params;

	return (*this);
}

//---------------------------------------------------------
int QEvolutionaryAlgorithm::setParameters (const QEAParameters& argParams) {
	_params = argParams;

	// create EA components
	_popul = _factory->createPopulation (_params.encoding);
	_selPopul = _factory->createPopulation (_params.encoding);
	_children = _factory->createPopulation (_params.encoding);
	_init = _factory->createInitializer (_params.encoding);
	_niching = _factory->createNichingOperator (_params.nichingType);
	_selection = _factory->createSelection (_params.selectionType);
	_populationSizer = _factory->createPopulationSizer (_params.populationSizerType);
	_xover = _factory->createCrossover (_params.encoding, _params.xType);
	_mutation = _factory->createMutation (_params.encoding, _params.mutationType);

	// initialize EA's components
	initializeComponents ();

	return Q_OK;
}

//---------------------------------------------------------
QRunSummary& QEvolutionaryAlgorithm::getSummary () {
	return _summary;
}

//---------------------------------------------------------
QPopulation* QEvolutionaryAlgorithm::getPopulation () const {
	return _popul;
}

//---------------------------------------------------------
int QEvolutionaryAlgorithm::initializeComponents () {
	QInitRand (_params.seed);

	// set parameters
	_popul->setParameters(_params);
	_selPopul->setParameters (_params);
	_children->setParameters (_params);
	_init->setParameters (_params);
	_selection->setParameters (_params);
	_selection->setPopulations (_popul, _selPopul);
	_populationSizer->setParameters (_params);
	_populationSizer->setPopulation (_children);
	if (_xover != NULL) {
		// since not all EAs use xover
		_xover->setParameters (_params);
		_xover->setParentSelector (_factory->createParentSelector(QPST_RANDOM));
		_xover->setPopulations (_selPopul, _children);
	}
	_mutation->setParameters (_params);

	if (_params.nichingType != QNT_UNKNOWN) {	// since niching is not obligatory operator
		_niching->setParameters (_params);
		_niching->setPopulation (_popul);

		if (_params.nichingType == QNT_SHARING) {
			((QFitnessSharing*)_niching)->setSigma (_params.sharingSigma);
		}
	}

	if (_params.populationSizerType == QPSZT_SIMPLE || _params.populationSizerType == QPSZT_SIMPLE_FIBONACCI) {
		((QSimplePopulationSizer*)_populationSizer)->setSummary (_summary);
		((QSimplePopulationSizer*)_populationSizer)->reset ();
	}

	_init->setPopulation (_popul);
	_init->operate ();
	_init->setPopulation (_selPopul);
	_init->operate ();
	_init->setPopulation (_children);
	_init->operate ();
	_children->clearPopulation ();

	_summary.means.assign (_params.generationsNumber, 0);
	_summary.devs.assign (_params.generationsNumber, 0);
	_summary.bests.assign (_params.generationsNumber, 0);
	_summary.worsts.assign (_params.generationsNumber, 0);
	_summary.popSizes.assign (_params.generationsNumber, _params.populationSize);
	if (_summary.currentBest != NULL) {
		delete _summary.currentBest;
	}

	return Q_OK;
}

//---------------------------------------------------------
int QEvolutionaryAlgorithm::evaluatePopulation () {
	int i;
	__VECTOR_DOUBLE ers;
	double mean, dev;//, minFitness = 100000000;

	if (_params.nichingType != QNT_CROWDING_DETERMINISTIC) {

		ers.resize (_popul->getSize());
		for (i=0; i<_popul->getSize(); i++) {
			_params.problem->evaluateIndividual (_popul->getIndividual (i));
			_summary.feCount++;
			ers [i] = _popul->getIndividualFitnessValue (i);
			if (ers [i] < _params.errorLevel && _summary.firstHitFECount == 0) {
				_summary.firstHitClock = clock ();
				_summary.firstHitFECount = _summary.feCount;
				_summary.firstHitGeneration = _summary.generation;
				_summary.firstHitTime = (double) (_summary.firstHitClock - _summary.startClock) / CLOCKS_PER_SEC;
				_summary.firstHitSolution = _popul->getIndividual (i)->clone ();
			}
		}

		// write statistics
		qstats (ers, mean, dev);
		_summary.means [_summary.generation] = mean;
		_summary.devs [_summary.generation] = dev;
		_summary.bests [_summary.generation] = (*min_element( ers.begin(), ers.end() ));
		_summary.worsts [_summary.generation] = (*max_element( ers.begin(), ers.end() ));
		if (_summary.currentBest != NULL) {
			delete _summary.currentBest;
		}
		if (_popul->getMinFitnessIndividual (0) != NULL) {
			_summary.currentBest = _popul->getMinFitnessIndividual (0)->clone ();
		} else {
			cout<<"?";
			_fout<<"strange results:\n";
			_popul->printPopulation (_fout);
			_fout<<"\n";
		}

		if (_summary.generation > 0 && _summary.bests [_summary.generation] == _summary.worsts [_summary.generation] && _summary.convergenceGeneration == 0)	{
			_summary.convergenceGeneration = _summary.generation + 1;
		}

		if (_params.nichingType == QNT_SHARING) {
			_niching->operate ();
		}

		if (_params.untilFirstHit && _summary.firstHitFECount != 0) {
			return Q_FAILURE;
		}

	} else {
		// do not evaluate fitness since population was evaluated during niching
		// but collect necessary statistics

		ers.resize (_popul->getSize());
		if (_summary.generation != 0) {
			for (i=0; i<_popul->getSize(); i++) {
				_summary.feCount++;
				ers [i] = _popul->getIndividualFitnessValue (i);
				if (ers [i] < _params.errorLevel && _summary.firstHitFECount == 0) {
					_summary.firstHitClock = clock ();
					_summary.firstHitFECount = _summary.feCount;
					_summary.firstHitGeneration = _summary.generation;
					_summary.firstHitTime = (double) (_summary.firstHitClock - _summary.startClock) / CLOCKS_PER_SEC;
				}
			}
		} else {
			for (i=0; i<_popul->getSize(); i++) {
				_params.problem->evaluateIndividual (_popul->getIndividual (i));
				_summary.feCount++;
				ers [i] = _popul->getIndividualFitnessValue (i);
				if (ers [i] < _params.errorLevel && _summary.firstHitFECount == 0) {
					_summary.firstHitClock = clock ();
					_summary.firstHitFECount = _summary.feCount;
					_summary.firstHitGeneration = _summary.generation;
					_summary.firstHitTime = (double) (_summary.firstHitClock - _summary.startClock) / CLOCKS_PER_SEC;
				}
			}
		}


		// write statistics
		qstats (ers, mean, dev);
		_summary.means [_summary.generation] = mean;
		_summary.devs [_summary.generation] = dev;
		_summary.bests [_summary.generation] = (*min_element( ers.begin(), ers.end() ));
		_summary.worsts [_summary.generation] = (*max_element( ers.begin(), ers.end() ));
		if (_summary.currentBest != NULL) {
			delete _summary.currentBest;
		}
		_summary.currentBest = _popul->getMinFitnessIndividual (0)->clone ();

		if (_summary.generation > 0 && _summary.bests [_summary.generation] == _summary.worsts [_summary.generation] && _summary.convergenceGeneration == 0)	{
			_summary.convergenceGeneration = _summary.generation + 1;
		}

	}	// if (_params.nichingType != QNT_CROWDING_DETERMINISTIC)

	return Q_OK;
}

//---------------------------------------------------------
int QEvolutionaryAlgorithm::evaluatePopulationNoStats () {
	int i;
	__VECTOR_DOUBLE ers;

	if (_params.nichingType != QNT_CROWDING_DETERMINISTIC) {

		ers.resize (_popul->getSize());
//		cout<<"ers.size() = "<<ers.size()<<"\n";
		for (i=0; i<_popul->getSize(); i++) {
			_params.problem->evaluateIndividual (_popul->getIndividual (i));
			_summary.feCount++;
			ers [i] = _popul->getIndividualFitnessValue (i);
			if (ers [i] < _params.errorLevel && _summary.firstHitFECount == 0) {
				_summary.firstHitClock = clock ();
				_summary.firstHitFECount = _summary.feCount;
				_summary.firstHitGeneration = _summary.generation;
				_summary.firstHitTime = (double) (_summary.firstHitClock - _summary.startClock) / CLOCKS_PER_SEC;
			}
		}

		// write statistics
		if (_summary.currentBest != NULL) {
			delete _summary.currentBest;
		}
		_summary.currentBest = _popul->getMinFitnessIndividual (0)->clone ();

		//cout<<"_summary.generation = "<<_summary.generation<<"\n";
		cout<<"Final best fitness = "<<_summary.bests [_summary.generation-1]<<"\n";
		cout<<"Final worst fitness = "<<_summary.worsts [_summary.generation-1]<<"\n";
		cout<<"Convergence generation = "<<_summary.convergenceGeneration<<"\n";

		if (_summary.generation == _params.generationsNumber) {
			if (_summary.generation > 0 && (_summary.bests [_summary.generation-1] == _summary.worsts [_summary.generation-1]) && (_summary.convergenceGeneration == 0)) {
				_summary.convergenceGeneration = _summary.generation;
			}
		} else {
			if (_summary.generation > 0 && (_summary.bests [_summary.generation] == _summary.worsts [_summary.generation]) && (_summary.convergenceGeneration == 0)) {
				_summary.convergenceGeneration = _summary.generation + 1;
			}
		}

		return Q_OK;
	} else {
		// do not evaluate fitness since population was evaluated during niching
		// but collect necessary statistics

		ers.resize (_popul->getSize());
		if (_summary.generation != 0) {
			for (i=0; i<_popul->getSize(); i++) {
				_summary.feCount++;
				ers [i] = _popul->getIndividualFitnessValue (i);
				if (ers [i] < _params.errorLevel && _summary.firstHitFECount == 0) {
					_summary.firstHitClock = clock ();
					_summary.firstHitFECount = _summary.feCount;
					_summary.firstHitGeneration = _summary.generation;
					_summary.firstHitTime = (double) (_summary.firstHitClock - _summary.startClock) / CLOCKS_PER_SEC;
				}
			}
		} else {
			for (i=0; i<_popul->getSize(); i++) {
				_params.problem->evaluateIndividual (_popul->getIndividual (i));
				_summary.feCount++;
				ers [i] = _popul->getIndividualFitnessValue (i);
				if (ers [i] < _params.errorLevel && _summary.firstHitFECount == 0) {
					_summary.firstHitClock = clock ();
					_summary.firstHitFECount = _summary.feCount;
					_summary.firstHitGeneration = _summary.generation;
					_summary.firstHitTime = (double) (_summary.firstHitClock - _summary.startClock) / CLOCKS_PER_SEC;
				}
			}
		}

		if (_summary.currentBest != NULL) {
			delete _summary.currentBest;
		}
		_summary.currentBest = _popul->getMinFitnessIndividual (0)->clone ();

		if (_summary.generation == _params.generationsNumber) {
			if (_summary.generation > 0 && (_summary.bests [_summary.generation-1] == _summary.worsts [_summary.generation-1]) && (_summary.convergenceGeneration == 0)) {
				_summary.convergenceGeneration = _summary.generation;
			}
		} else {
			if (_summary.generation > 0 && (_summary.bests [_summary.generation] == _summary.worsts [_summary.generation]) && (_summary.convergenceGeneration == 0)) {
				_summary.convergenceGeneration = _summary.generation + 1;
			}
		}

	}	// if (_params.nichingType != QNT_CROWDING_DETERMINISTIC)

	return Q_OK;
}

//---------------------------------------------------------
int QEvolutionaryAlgorithm::selectParentIndividuals () {
	if (_params.nichingType != QNT_CROWDING_DETERMINISTIC) {
		_selection->operate ();

	} else {
		// shuffle population
		int i=0, size = _popul->getSize();
		_selPopul->clearPopulation ();
		for (i=0; i<size; i++) {
			// to make fast shuffling individuals are simply added at random
			// either to the top or bottom of the parental population
			if (QRandBool()) {
				_selPopul->addIndividual (_popul->getIndividual(i));
			} else {
				_selPopul->insertIndividual (0, _popul->getIndividual(i));
			}
		}
	}	// if (_params.nichingType != QNT_CROWDING_DETERMINISTIC)
	return Q_OK;
}

//---------------------------------------------------------
int QEvolutionaryAlgorithm::selectElite () {
	if (_params.eliteCount > 0) {
		if (_params.eliteCount < _selPopul->getSize ()) {
			QPopulation* tempPopul;
			int i, k;

			tempPopul = _factory->createPopulation (_selPopul->getEncodingType ());
			tempPopul->assign (_selPopul);
			for (k=0; k<_params.eliteCount; k++) {
				i = tempPopul->getMinFitnessIndividualIndex (0.0);

				if (i >= 0) {
					_children->addIndividual (tempPopul->getIndividual(i));
					tempPopul->deleteIndividual (i);
				} else {
					cerr<<"\n\n[QEvolutionaryAlgorithm::selectElite] error:\n"
						<<"\tUnable to perform elite selection\n\n";
					_fout<<"\n\n[QEvolutionaryAlgorithm::selectElite] error:\n"
						<<"\tUnable to perform elite selection\n\n";
					_popul->printPopulation (_fout);
					exit (0);
				}
			}

			_factory->deleteLastCreatedPopulation ();

			return Q_OK;
		} else {
			cerr<<"\n\n[QEvolutionaryAlgorithm::selectElite] error:\n"
				<<"\tNumber of elite individuals exceeds total number of individuals\n\n";
			exit (0);
		}	// if (_params.eliteCount < _selPopul->getSize ())
	}	// if (_params.eliteCount > 0)
	return Q_OK;
}

//---------------------------------------------------------
int QEvolutionaryAlgorithm::crossPopulation () {

	while ( !_populationSizer->populationReady() ) {
		//printf ("Population size %i", _children->getSize ());
		//fflush (stdout);
		_xover->operate ();
	}
	_populationSizer->operate ();

	return Q_OK;
}

//---------------------------------------------------------
int QEvolutionaryAlgorithm::mutatePopulation () {
	if (_params.nichingType != QNT_CROWDING_DETERMINISTIC) {
		int i;

		for (i=0; i<_children->getSize(); i++) {
			_mutation->setMutant (_children->getIndividual(i));
			_mutation->operate ();
		}
	}

	return Q_OK;
}

//---------------------------------------------------------
QMutationOperator* QEvolutionaryAlgorithm::getMutationOperator () const {
	return _mutation;
}

//---------------------------------------------------------
QNichingOperator* QEvolutionaryAlgorithm::getNichingOperator () const {
	return _niching;
}

//---------------------------------------------------------
QEAFactory* QEvolutionaryAlgorithm::getFactory () const {
	return _factory;
}

//---------------------------------------------------------
int QEvolutionaryAlgorithm::nextGeneration () {
	_popul->assign (_children);
	_children->clearPopulation ();
	return Q_OK;	//_popul->checkConvergence ();
}

//---------------------------------------------------------
int QEvolutionaryAlgorithm::finishWork () {
	_factory->deleteComponents ();
	_summary.reset ();
	return Q_OK;
}
