/*****************************************

ECWorkshop -- evolutionary computation
toolkit library

Web: http://qai.narod.ru/ecw/
Mailto: qai@mail.ru

Copyright (C) 2007 by Yury Tsoy


This program is free software; you can
redistribute it and/or modify it under the
terms of the GNU General Public License as
published by the Free Software Foundation;
either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that
it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU General Public License
for more details.

You should have received a copy of the GNU
General Public License along with this
program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330,
Boston, MA  02111-1307  USA

*/

#include "qsarn2.h"
#include "../problems/qannproblem.h"

#include <algorithm>

using namespace ECWorkshop;

QSARN2::QSARN2 (const QSARN2& argEA) : QEvolutionaryAlgorithm(argEA) {
	//_net = argEA._net->clone ();
}

QSARN2& QSARN2::operator= (const QSARN2& argEA) {
	// no assignment is made
	//_net = argEA._net->clone ();
	return *this;
}

QSARN2::QSARN2 (void) {
	_fout.open ("sarn2.log");
	_net = new QRNIndividual ();
	_mut = new QRNMutationBasic (this);
}

QSARN2::~QSARN2 (void) {
	_fout.close ();
	if (_net != NULL) {
		delete _net;
	}
	if (_mut != NULL) {
		delete _mut;
	}
}

// int setParameters (const QGAParameters& argParams)
//	Sets parameters of the EA and makes necessary initializations
//	Returns Q_OK on success
int QSARN2::setParameters (const QEAParameters& argParams) {
	this->QEvolutionaryAlgorithm::setParameters (argParams);

	// init mutation
	_mut->setParameters (argParams);

	// init RN
	QRNGene* tempGene;
	QSignal tempSignal;
	int i, k, chrSize;

	_net->setParameters (&argParams);
	chrSize = _net->getSize ();

	for (i=0; i<chrSize; i++) {
		tempGene = ( QRNGene* )_net->getGene(i);

		// set number of inputs and outputs
		tempGene->setInputsNumber( _params.inputs );
		tempGene->setOutputsNumber( _params.outputs );

		// set random input signal from range [-0.1; +0.1] for the node
		for (k=1; k<=_params.inputs; k++) {
			tempSignal._id = k;
			tempSignal._amount = 0.0;
			tempGene->setInput (k-1, tempSignal);
			// set random input weight
			tempGene->setInputWeight (k-1, 0.2*QRandN() - 0.1);
		}
		// set output signals for the node
		for (; k<=(_params.inputs + _params.outputs); k++) {
//				cout<<"k = "<<k<<"\n";
			tempSignal._id = k;
			tempSignal._amount = 0.0;
			tempGene->setOutput ((k - _params.inputs-1), tempSignal);
			// set output weight
			tempGene->setOutputWeight ((k - _params.inputs-1), 1.0);
		}
		tempGene->setBias (0.2*QRandN() - 0.1);
		tempGene->setParamA (10.0*QRandGauss());
	}

	return Q_OK;
}

// int run ()
//	Performs the main routine of evolutionary algorithm/model
//	Returns Q_OK on success
int QSARN2::run () {
	if (_params.generationsNumber > 0) {
		int i;

		_summary.reset ();
		_summary.startClock = clock ();

		initializeComponents ();

		_fout<<"\nStarting a new run:\n"
			<<"-------------------------------\n";
		_popul->printPopulation (_fout);
		_fout.flush ();

		for (i=0; i<_params.generationsNumber; i++) {

			_summary.popSizes[ _summary.generation ] = _popul->getSize();
			//cout<<"evaluatePopulation\n";
			//cout.flush ();
			if (evaluatePopulation () == Q_OK) {
				//cout<<"selectParentIndividuals\n";
				//cout.flush ();
				selectParentIndividuals ();
	//			cout<<"selectElite\n";
	//			cout.flush ();
				selectElite ();
				//cout<<"crossPopulation\n";
				//cout.flush ();
				crossPopulation ();
				//cout<<"mutatePopulation\n";
				//cout.flush ();
				mutatePopulation ();
				//cout<<"nextGeneration\n";
				//cout.flush ();
				nextGeneration ();

				_summary.generation++;

				if (! (_summary.generation % 10)) {
					printf(".");
				}
			} else {	// first hit event occured with the [QEAParameters::untilFirstHit] flag set
				break;
			}
		}
		printf ("\nFinishing the run...");

		_summary.finishClock = clock ();
		_summary.time = (double) (_summary.finishClock - _summary.startClock) / CLOCKS_PER_SEC;

		printf("done\n");

		evaluatePopulationNoStats ();
		_popul->printPopulation (_fout);

		// test found solution or current best individual
		if (_summary.firstHitSolution != NULL) {
			((QANNProblem*)_params.problem)->testIndividual (_summary.firstHitSolution);
		} else {
			((QANNProblem*)_params.problem)->testIndividual (_summary.currentBest);
		}

		_fout<<"Run duration: "<<_summary.time<<" seconds\n";

		return Q_OK;
	} else {
		cerr<<"\n\n[QSARN2::run] error:\n"
			<<"\tGenerations number is undefined! (_params.generationsNumber = "<<_params.generationsNumber<<")\n\n";
		return Q_FAILURE;
	}
	return Q_OK;
}

// QEvolutionaryAlgorithm* clone ()
//	Makes copy of the existing object
QEvolutionaryAlgorithm* QSARN2::clone () {
	return new QSARN2 (*this);
}

// int evaluatePopulation
// Overloaded method
int QSARN2::evaluatePopulation () {
	int i, j, popSize = _popul->getSize (), temp;
	__VECTOR_DOUBLE ers, ers_1, mparam;
	double mean, dev, maxEr;

	// create temporal population of [RNs]
	// from the mutations of the current [_net]
	QPopulation* populRN = _factory->createPopulation(QET_REGULATORY_NETWORK);
	QRCGene tempGene;
	for (i=0; i<popSize; ++i) {
		// set i-th mutation parameters vector
		_popul->getIndividual(i)->toVectorDouble (mparam);
		for (j=0; j<MUTATION_PARAMETERS; ++j) {
			_net->setMutationParam (j, mparam[j]);
		}
		populRN->addIndividual (_net);
		//for (j=0, temp=0; j<MUTATION_PARAMETERS; ++j) {
		//	temp += ((QRNIndividual*)populRN->getIndividual(i))->readHistory(j);
		//}
		//if (_summary.generation > 0 && temp == 0) {
		//	cout<<"no history!\n";
		//}
		_mut->setMutant (populRN->getIndividual(i));
		_mut->operate ();

		// write new mutation parameters (after mutation)
		for (j=0; j<MUTATION_PARAMETERS; ++j) {
			tempGene.setValue (((QRNIndividual*)populRN->getIndividual(i))->getMutationParam(j));
			_popul->getIndividual(i)->setGene (j, &tempGene);
		}
	}

	if (_params.nichingType != QNT_CROWDING_DETERMINISTIC) {

		ers.resize (popSize);
		for (i=0; i<popSize; i++) {
			_params.problem->evaluateIndividual (populRN->getIndividual (i));
			_popul->getIndividual(i)->setFitness (populRN->getIndividual (i)->getFitness());	// copy fitness
			_summary.feCount++;
			ers [i] = _popul->getIndividualFitnessValue (i);
			if (ers [i] < _params.errorLevel && _summary.firstHitFECount == 0) {
				_summary.firstHitClock = clock ();
				_summary.firstHitFECount = _summary.feCount;
				_summary.firstHitGeneration = _summary.generation;
				_summary.firstHitTime = (double) (_summary.firstHitClock - _summary.startClock) / CLOCKS_PER_SEC;
				_summary.firstHitSolution = populRN->getIndividual (i)->clone ();
			}
		}

		// apply changes to the current [RN] using roulette wheel
		// *invert* [ers] values to get fitness (not error)
		maxEr = (*max_element(ers.begin(), ers.end()));
		ers_1.resize (popSize);
		for (i=0; i<popSize; ++i) {
			ers_1[i] = maxEr - ers[i];
		}
		i = roulette (ers_1, 1);
		_net->assign (populRN->getIndividual (i));
		//for (j=0, temp=0; j<MUTATION_PARAMETERS; ++j) {
		//	temp += _net->readHistory(j);
		//}
		//if (_summary.generation > 0 && temp == 0) {
		//	cout<<"no history 2!\n";
		//}

		// write statistics
		qstats (ers, mean, dev);
		_summary.means [_summary.generation] = mean;
		_summary.devs [_summary.generation] = dev;
		_summary.bests [_summary.generation] = (*min_element( ers.begin(), ers.end() ));
		_summary.worsts [_summary.generation] = (*max_element( ers.begin(), ers.end() ));
		if (populRN->getMinFitnessIndividual (0) != NULL) {
			if (_summary.currentBest != NULL) {
				delete _summary.currentBest;
			}
			_summary.currentBest = populRN->getMinFitnessIndividual (0)->clone ();
		} else {
			cout<<"?";
			_fout<<"strange results:\n";
			_popul->printPopulation (_fout);
			_fout<<"\n";
		}

		if (_summary.generation > 0 && _summary.bests [_summary.generation] == _summary.worsts [_summary.generation] && _summary.convergenceGeneration == 0)	{
			_summary.convergenceGeneration = _summary.generation + 1;
		}

		if (_params.nichingType == QNT_SHARING) {
			_niching->operate ();
		}

		if (_params.untilFirstHit && _summary.firstHitFECount != 0) {
			return Q_FAILURE;
		}

	} else {
		// do not evaluate fitness since population was evaluated during niching
		// but collect necessary statistics

		ers.resize (_popul->getSize());
		if (_summary.generation != 0) {
			for (i=0; i<_popul->getSize(); i++) {
				_summary.feCount++;
				ers [i] = _popul->getIndividualFitnessValue (i);
				if (ers [i] < _params.errorLevel && _summary.firstHitFECount == 0) {
					_summary.firstHitClock = clock ();
					_summary.firstHitFECount = _summary.feCount;
					_summary.firstHitGeneration = _summary.generation;
					_summary.firstHitTime = (double) (_summary.firstHitClock - _summary.startClock) / CLOCKS_PER_SEC;
					_summary.firstHitSolution = populRN->getIndividual (i)->clone ();
				}
			}
		} else {
			for (i=0; i<_popul->getSize(); i++) {
				_params.problem->evaluateIndividual (populRN->getIndividual (i));
				_popul->getIndividual (i)->setFitness (populRN->getIndividual(i)->getFitness());
				_summary.feCount++;
				ers [i] = _popul->getIndividualFitnessValue (i);
				if (ers [i] < _params.errorLevel && _summary.firstHitFECount == 0) {
					_summary.firstHitClock = clock ();
					_summary.firstHitFECount = _summary.feCount;
					_summary.firstHitGeneration = _summary.generation;
					_summary.firstHitTime = (double) (_summary.firstHitClock - _summary.startClock) / CLOCKS_PER_SEC;
					_summary.firstHitSolution = populRN->getIndividual (i)->clone ();
				}
			}
		}

		// write statistics
		qstats (ers, mean, dev);
		_summary.means [_summary.generation] = mean;
		_summary.devs [_summary.generation] = dev;
		_summary.bests [_summary.generation] = (*min_element( ers.begin(), ers.end() ));
		_summary.worsts [_summary.generation] = (*max_element( ers.begin(), ers.end() ));
		if (_summary.currentBest != NULL) {
			delete _summary.currentBest;
		}
		_summary.currentBest = populRN->getMinFitnessIndividual (0)->clone ();

		if (_summary.generation > 0 && _summary.bests [_summary.generation] == _summary.worsts [_summary.generation] && _summary.convergenceGeneration == 0)	{
			_summary.convergenceGeneration = _summary.generation + 1;
		}

	}	// if (_params.nichingType != QNT_CROWDING_DETERMINISTIC)

	// remove temporal population
	_factory->deleteLastCreatedPopulation ();
	populRN = NULL;

	return Q_OK;
}

int QSARN2::evaluatePopulationNoStats () {
	int i, j, popSize = _popul->getSize();
	__VECTOR_DOUBLE ers, mparam;

	// create temporal population of [RNs]
	// from the mutations of the current [_net]
	QPopulation* populRN = _factory->createPopulation(QET_REGULATORY_NETWORK);
	for (i=0; i<popSize; ++i) {
		// set i-th mutation parameters vector
		_popul->getIndividual(i)->toVectorDouble (mparam);
		for (j=0; j<MUTATION_PARAMETERS; ++j) {
			_net->setMutationParam (j, mparam[j]);
		}
		populRN->addIndividual (_net);
		_mut->setMutant (populRN->getIndividual(i));
		_mut->operate ();
	}

	if (_params.nichingType != QNT_CROWDING_DETERMINISTIC) {

		ers.resize (_popul->getSize());
//		cout<<"ers.size() = "<<ers.size()<<"\n";
		for (i=0; i<_popul->getSize(); i++) {
			_params.problem->evaluateIndividual (populRN->getIndividual (i));
			_popul->getIndividual(i)->setFitness (populRN->getIndividual(i)->getFitness());
			_summary.feCount++;
			ers [i] = _popul->getIndividualFitnessValue (i);
			if (ers [i] < _params.errorLevel && _summary.firstHitFECount == 0) {
				_summary.firstHitClock = clock ();
				_summary.firstHitFECount = _summary.feCount;
				_summary.firstHitGeneration = _summary.generation;
				_summary.firstHitTime = (double) (_summary.firstHitClock - _summary.startClock) / CLOCKS_PER_SEC;
				_summary.firstHitSolution = populRN->getIndividual(i);
			}
		}

		// write statistics
		if (_summary.currentBest != NULL) {
			delete _summary.currentBest;
		}
		_summary.currentBest = populRN->getMinFitnessIndividual (0)->clone ();

		//cout<<"_summary.generation = "<<_summary.generation<<"\n";
		cout<<"Final best fitness = "<<_summary.bests [_summary.generation-1]<<"\n";
		cout<<"Final worst fitness = "<<_summary.worsts [_summary.generation-1]<<"\n";
		cout<<"Convergence generation = "<<_summary.convergenceGeneration<<"\n";

		if (_summary.generation == _params.generationsNumber) {
			if (_summary.generation > 0 && (_summary.bests [_summary.generation-1] == _summary.worsts [_summary.generation-1]) && (_summary.convergenceGeneration == 0)) {
				_summary.convergenceGeneration = _summary.generation;
			}
		} else {
			if (_summary.generation > 0 && (_summary.bests [_summary.generation] == _summary.worsts [_summary.generation]) && (_summary.convergenceGeneration == 0)) {
				_summary.convergenceGeneration = _summary.generation + 1;
			}
		}

		return Q_OK;
	} else {
		// do not evaluate fitness since population was evaluated during niching
		// but collect necessary statistics

		ers.resize (_popul->getSize());
		if (_summary.generation != 0) {
			for (i=0; i<_popul->getSize(); i++) {
				_summary.feCount++;
				ers [i] = _popul->getIndividualFitnessValue (i);
				if (ers [i] < _params.errorLevel && _summary.firstHitFECount == 0) {
					_summary.firstHitClock = clock ();
					_summary.firstHitFECount = _summary.feCount;
					_summary.firstHitGeneration = _summary.generation;
					_summary.firstHitTime = (double) (_summary.firstHitClock - _summary.startClock) / CLOCKS_PER_SEC;
				}
			}
		} else {
			for (i=0; i<_popul->getSize(); i++) {
				_params.problem->evaluateIndividual (_popul->getIndividual (i));
				_summary.feCount++;
				ers [i] = _popul->getIndividualFitnessValue (i);
				if (ers [i] < _params.errorLevel && _summary.firstHitFECount == 0) {
					_summary.firstHitClock = clock ();
					_summary.firstHitFECount = _summary.feCount;
					_summary.firstHitGeneration = _summary.generation;
					_summary.firstHitTime = (double) (_summary.firstHitClock - _summary.startClock) / CLOCKS_PER_SEC;
				}
			}
		}

		if (_summary.currentBest != NULL) {
			delete _summary.currentBest;
		}
		_summary.currentBest = _popul->getMinFitnessIndividual (0)->clone ();

		if (_summary.generation == _params.generationsNumber) {
			if (_summary.generation > 0 && (_summary.bests [_summary.generation-1] == _summary.worsts [_summary.generation-1]) && (_summary.convergenceGeneration == 0)) {
				_summary.convergenceGeneration = _summary.generation;
			}
		} else {
			if (_summary.generation > 0 && (_summary.bests [_summary.generation] == _summary.worsts [_summary.generation]) && (_summary.convergenceGeneration == 0)) {
				_summary.convergenceGeneration = _summary.generation + 1;
			}
		}

	}	// if (_params.nichingType != QNT_CROWDING_DETERMINISTIC)

	return Q_OK;
}

// int printResults (__VECTOR_RUN_SUMMARY& argS, ostream& out)
//	Prints algorithm specific results into the passed output stream
int QSARN2::printResults (__VECTOR_RUN_SUMMARY& argS, ostream& out) {
	int sols;	// solutions number
	int i, j, size = (int)argS.size(), temp, size2;
	double mean, dev, meanC, devC;
	__VECTOR_INT tempV, tempD;
	__VECTOR_DOUBLE mutS, mutC;

	// calculate number of solutions
	// and average number of nodes pur population of solutions
	// and population of candidates ([meanC] and [devC])
	sols = 0;
	mean = dev = 0.0;
	meanC = devC = 0.0;
	for (i=0; i<size; ++i) {
		if (argS[i].firstHitFECount != 0) {
			// calculate stats for solutions
			++sols;
			temp = argS[i].firstHitSolution->getSize();
			mean += temp;
			dev += temp*temp;
		} else {
			// calculate stats for candidates
			temp = argS[i].currentBest->getSize();
			meanC += temp;
			devC += temp*temp;
		}
	}
	mean /= (double)sols;
	dev = dev/(double)sols - mean*mean;
	dev *= (double)sols/(sols-1);
	if (sols < size) {
		meanC /= (double)(size-sols);
		devC = devC/(double)(size-sols) - meanC*meanC;
		devC *= (double)(size-sols)/((size-sols)-1);
	}

	// print average number of nodes
	out<<"> Solutions summary ("<<sols<<"):\n"
		<<"Average number of nodes:\t"<<mean<<"\n"
		<<"Deviation of the number of nodes:\t"<<dev<<"\n";
	if (sols < size) {
		out<<"> Candidate-solutions summary ("<<(size-sols)<<"):\n"
			<<"Average number of nodes:\t"<<meanC<<"\n"
			<<"Deviation of the number of nodes:\t"<<devC<<"\n";
	}

	// calculate data about distribution of nodes degrees
	mean = dev = 0.0;
	for (i=0; i<size; ++i) {
		if (argS[i].firstHitTime != 0) {
			((QRNIndividual*)(argS[i].firstHitSolution))->calculateDegrees (tempV);

			size2 = (int)tempV.size ();
			for (j=0; j<size2; ++j) {
				while ((int)tempD.size() <= tempV[j]) {
					tempD.push_back (0);
				}
				tempD[ tempV[j] ]++;
			}

			// calculate average and deviation for nodes degrees
			qstats (tempV, meanC, devC);
			mean += meanC;
			dev += devC;
		}
	}
	if (sols > 0) {
		mean /= (double)sols;
		dev /= (double)sols;
		out<<"> Distribution of nodes degrees over the population of solutions:\t";
		size2 = (int)tempD.size ();
		for (i=0; i<size2; ++i) {
			out<<" "<<tempD[i];
		}
		out<<"\n";
		out<<"\tMean nodes degree:\t"<<mean<<"\n"
			<<"\tAveraged deviation of the nodes degree:\t"<<dev<<"\n";
	}

	// calculate averaged vector of mutation parameters
	mutS.assign (MUTATION_PARAMETERS, 0.0);
	mutC.assign (MUTATION_PARAMETERS, 0.0);
	for (i=0; i<size; ++i) {
		if (argS[i].firstHitTime != 0) {
			// calculate history for solutions
			for (j=0; j<MUTATION_PARAMETERS; ++j) {
				mutS[j] += ((QRNIndividual*)(argS[i].firstHitSolution))->getMutationParam (j);
			}
		} else {
			// calculate history for candidates
			for (j=0; j<MUTATION_PARAMETERS; ++j) {
				mutC[j] += ((QRNIndividual*)(argS[i].currentBest))->getMutationParam (j);
			}
		}
	}
	if (sols > 0) {
		out<<"> Averaged vector of mutation parameters for the solutions:\t";
		for (i=0; i<MUTATION_PARAMETERS; ++i) {
			out<<mutS[i]/(double)sols<<" ";
		}
		out<<"\n";
	}
	if (sols < size) {
		out<<"> Averaged vector of mutation parameters for the candidates:\t";
		for (i=0; i<MUTATION_PARAMETERS; ++i) {
			out<<mutC[i]/(double)(size-sols)<<" ";
		}
		out<<"\n";
	}

	// calculate cumulative history events
	tempV.assign (MUTATION_PARAMETERS, 0);
	tempD.assign (MUTATION_PARAMETERS, 0);
	for (i=0; i<size; ++i) {
		if (argS[i].firstHitTime != 0) {
			// calculate history for solutions
			for (j=0; j<MUTATION_PARAMETERS; ++j) {
				tempV[j] += ((QRNIndividual*)(argS[i].firstHitSolution))->readHistory (j);
			}
		} else {
			// calculate history for candidates
			for (j=0; j<MUTATION_PARAMETERS; ++j) {
				tempD[j] += ((QRNIndividual*)(argS[i].currentBest))->readHistory (j);
			}
		}
	}
	out<<"> Cumulative history events counters for the solutions:\t";
	for (i=0; i<MUTATION_PARAMETERS; ++i) {
		out<<tempV[i]<<" ";
	}
	out<<"\n";
	if (sols < size) {
		out<<"> Cumulative history events counters for the candidates:\t";
		for (i=0; i<MUTATION_PARAMETERS; ++i) {
			out<<tempD[i]<<" ";
		}
		out<<"\n";
	}

	return Q_OK;
}

// int printName (ostream& out)
//	Prints algorithm name into the specified output stream
int QSARN2::printName (ostream& out) {
	out<<"Self-adaptive regulatory network (indirect evolution)\n";
	return Q_OK;
}


