/*****************************************

ECWorkshop -- evolutionary computation
toolkit library

Web: http://qai.narod.ru/ecw/
Mailto: qai@mail.ru

Copyright (C) 2007 by Yury Tsoy


This program is free software; you can
redistribute it and/or modify it under the
terms of the GNU General Public License as
published by the Free Software Foundation;
either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that
it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU General Public License
for more details.

You should have received a copy of the GNU
General Public License along with this
program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330,
Boston, MA  02111-1307  USA

*/

#ifndef	Q_REGULATORYNETWORK_GENE_H
#define	Q_REGULATORYNETWORK_GENE_H

#include "qgene.h"

#include "../../qtools/qtools.h"
#include "../include/qec.h"

namespace ECWorkshop {

// Baseline enums

enum QRNNodeType {QRNT_UNKNOWN=0, QRNT_THRESHOLD, QRNT_SIGMOID, QRNT_LINEAR, QRNT_SIN, QRNT_SIZE};


/////////////////////////////////////////////////////////////
//
//	Struct QSignal
//
/////////////////////////////////////////////////////////////

struct QSignal {
	int _id;	// id's for signals should begin from 1.
				// id = 0 is reserved for 'null' signal
	double _amount;

	QSignal () : _id(0), _amount(0.0) {}

	bool operator== (const QSignal& argP) const {
		if (_id == argP._id && _amount == argP._amount) {
			return true;
		} else {
			return false;
		}
	}

	bool operator!= (const QSignal& argP) const {
		if (_id == argP._id && _amount == argP._amount) {
			return false;
		} else {
			return true;
		}
	}
};

typedef		vector<QSignal>	__VECTOR_SIGNAL;

/////////////////////////////////////////////////////////////
//
//	Class QRNGene
//
// Base class for regulatory network genes representation
//
/////////////////////////////////////////////////////////////

class QRNGene : public QGene {
protected:

	QRNNodeType _type;	// node type

	__VECTOR_SIGNAL _input;		// input signals array
	__VECTOR_SIGNAL _output;	// output signals array
	__VECTOR_DOUBLE _weight;	// weights of usage of input signals
	__VECTOR_DOUBLE _weightOuts;	// weights of usage of input signals

	double _bias;
	double _a;	// parameter of the transfer function
	bool _enabled;

public:

	QRNGene (void);
	QRNGene (const QRNGene& argGene);
	virtual ~QRNGene (void);

	virtual QRNGene& operator= (const QRNGene& argGene);

	virtual QGene* clone () const;
	virtual QEncodingType getEncodingType () const;
	virtual bool equalsTo (const QGene* argGene) const;
	virtual int assign (const QGene* argGene);

	// set number of inputs
	virtual int setInputsNumber (const int argN);
	int getInputsNumber () const;

	// set number of outputs
	virtual int setOutputsNumber (const int argN);
	int getOutputsNumber () const;

	// get product protein by its id or index
	virtual QSignal* getOutput (const int argIndex);
	virtual const QSignal* getOutput (const int argIndex) const;
	virtual QSignal* getOutputByID (const int argID);

	virtual QSignal* getInput (const int argIndex);
	virtual const QSignal* getInput (const int argIndex) const;

	// set amount of input protein by its id (the 1st and 2nd methods) or index (the last method)
	virtual int setInput (const QSignal& argValue);
	virtual int setInputByID (const int argID, const double argAmount);
	virtual int setInput (const int argIndex, const double argAmount);
	virtual int setInput (const int argIndex, const QSignal& argValue);
	virtual int setOutput (const int argIndex, const QSignal& argValue);

	// set input/output signals (with replacement of existing ones)
	virtual int setInputs (const __VECTOR_SIGNAL& argIns);
	virtual int setOutputs (const __VECTOR_SIGNAL& argOut);

	// get max ids for input and output signals
	virtual int getMaxInputID () const;
	virtual int getMaxOutputID () const;

	// add/delete signals methods
	virtual int addInputSignal (const QSignal* argS, const double argW);
	virtual int deleteInputSignalByID (const int argID);
	virtual int addOutputSignal (const QSignal* argS, const double argW);
	virtual int deleteOutputSignalByID (const int argID);

	// check available signals
	bool hasInput (const int id);
	bool hasOutput (const int id);

	// set weight for the desired input signal
	int setInputWeight (const int argIndex, double argW);
	double getInputWeight (const int argIndex) const;
	// set weight for the desired output signal
	int setOutputWeight (const int argIndex, double argW);
	double getOutputWeight (const int argIndex) const;

	int changeInputID (const int oldID, const int newID);
	int changeOutputID (const int oldID, const int newID);

	int calculate (const __VECTOR_DOUBLE& curState, __VECTOR_DOUBLE& newState);

	bool isEnabled () const;
	int enable (bool argStatus);

	double getBias () const;
	int setBias (const double argB);
	int setType (const QRNNodeType argT);
	QRNNodeType getType () const;

	double getParamA () const;
	int setParamA (const double argA);

	virtual int print2Stream (ostream& out) const;
};

typedef		vector<QRNGene>	__VECTOR_RN_GENE;

};

#endif
