/*****************************************

ECWorkshop -- evolutionary computation
toolkit library

Web: http://qai.narod.ru/ecw/
Mailto: qai@mail.ru

Copyright (C) 2007 by Yury Tsoy


This program is free software; you can
redistribute it and/or modify it under the
terms of the GNU General Public License as
published by the Free Software Foundation;
either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that
it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU General Public License
for more details.

You should have received a copy of the GNU
General Public License along with this
program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330,
Boston, MA  02111-1307  USA

*/

#include "qcrossover.h"

#include "../../eas/include.h"
#include "../niching/qcrowding.h"

using namespace ECWorkshop;

///////////////////////////////////////////////////////////
//
//		QCrossoverOperator class
//
///////////////////////////////////////////////////////////

QCrossoverOperator::QCrossoverOperator (void)
: _pSelector(NULL), _children(NULL) {
}

QCrossoverOperator::QCrossoverOperator (const QEvolutionaryAlgorithm* owner)
: QOperator(owner), _pSelector(NULL), _children(NULL) {
}

QCrossoverOperator::QCrossoverOperator (const QCrossoverOperator& argX)
: QOperator(argX), _pSelector(argX._pSelector), _children(argX._children) {
}

QCrossoverOperator::~QCrossoverOperator (void) {
	//_fout.close ();
}

QCrossoverOperator& QCrossoverOperator::operator= (const QCrossoverOperator& argX) {
	this->QOperator::operator = (argX);

	_pSelector = argX._pSelector;
    _children = argX._children;

	return (*this);
}

//int QCrossoverOperator::setParameters (const QEAParameters& argParams) {
//	QOperator::setParameters (argParams);
//}

int QCrossoverOperator::setPopulations (const QPopulation* argParents, QPopulation* argChildren) {
	if (_pSelector != NULL) {
		_pSelector->setPopulation (argParents);
		_children = argChildren;
	} else {
		cerr<<"\nQCrossoverOperator::setPopulations error:\n\t"
			<<"QParentSelector object is NULL!\n\n";
		exit (0);
	}

	return Q_OK;
}

int QCrossoverOperator::setParentSelector (const QParentSelector* argSelector) {
	_pSelector = argSelector;

	return Q_OK;
}

int QCrossoverOperator::operate () {
	if (_params->parentsNumber == 2) {	// for now only crossover for 2 parents is supported
		int i;
		__VECTOR_CONST_INDIVIDUAL_PTR parents;
		__VECTOR_INDIVIDUAL_PTR children;
		int genesNumber = _params->chromosomeLength, geneSize = _params->geneSize, geneMask = (int)splitMask[geneSize];
		int totalBits = _params->chromosomeLength * _params->geneSize;

		// select parents
		parents.resize (_params->parentsNumber);
		if (_params->nichingType != QNT_CROWDING_DETERMINISTIC) {
			parents[0] = _pSelector->selectParent ();
			parents[1] = _pSelector->selectParent ();
		} else {
			// pick pairs of parent individuals one by one
			parents[0] = _owner->getPopulation()->getIndividual(_children->getSize() - _params->eliteCount);
			parents[1] = _owner->getPopulation()->getIndividual(_children->getSize() - _params->eliteCount + 1);
		}

		// prepare children chromosomes
		children.resize (_params->childrenNumber);
		for (i=0; i<_params->childrenNumber; i++) {
			_children->addIndividual ();
			children[i] = _children->getIndividual ( _children->getSize() - 1 );
		}

		if (_params->nichingType != QNT_CROWDING_DETERMINISTIC) {
			if (_params->xRate >= QRandN()) {

				cross (&parents, &children);

			} else {	// xRate < than QRandN ()

				// no crossing, simply copies of parent individuals pass into the next generation
				children[0]->assign (parents[0]);
				children[1]->assign (parents[1]);
			}	// if (_params->xRate > QRandN())

		} else {
			// [QNT_CROWDING_DETERMINISTIC] case
			cross (&parents, &children);

			//printf ("%i, %i:", _children->getSize()-1, _children->getSize());
			QDeterministicCrowding* tempNiching = (QDeterministicCrowding*)_owner->getNichingOperator ();
			tempNiching->setParents ((QIndividual*)parents[0], (QIndividual*)parents[1]);
			tempNiching->setChildren ((QIndividual*)children[0], (QIndividual*)children[1]);
			tempNiching->operate ();
			//printf ("\n");
		}	// if (_params->nichingType != QNT_CROWDING_DETERMINISTIC)

		// **************************************************
		// if clones removal is on
		if (_params->removeClones) {
			int removeFlag = 0;	// flag to determine how many child chromosomes should be removed
								// 0 -- if none
								// 1 -- if the first child
								// 2 -- if the second child
								// 3 -- if both
			// check whether two last added individuals are new
			for (i=0; i<_children->getSize()-2; i++) {
				if (!(removeFlag & 0x1) && _children->getIndividual (i)->equalsTo( children[0] )) {
					removeFlag |= 0x1;
				}
				if (!(removeFlag & 0x2) && _children->getIndividual (i)->equalsTo( children[1] )) {
					removeFlag |= 0x2;
				}
				if (removeFlag == 3) {
					break;
				}
			}
			// remove clones
			if (removeFlag & 0x1) {
				// remove child before the last one
				_children->deleteIndividual (_children->getSize()-2);
			}
			if (removeFlag & 0x2) {
				// remove the last child
				_children->deleteIndividual (_children->getSize()-1);
			}

		}	// if (_params->removeClones)

	} else {
		cerr<<"!!! QIntegerCrossover2::operate warning:\t"
			<<"Parents number is unequal to 2. No crossing is performed !!!\n";
	}	// if (_params->parentsNumber == 2)
	return Q_OK;
}
