/*****************************************

ECWorkshop -- evolutionary computation
toolkit library

Web: http://qai.narod.ru/ecw/
Mailto: qai@mail.ru

Copyright (C) 2007 by Yury Tsoy


This program is free software; you can
redistribute it and/or modify it under the
terms of the GNU General Public License as
published by the Free Software Foundation;
either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that
it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU General Public License
for more details.

You should have received a copy of the GNU
General Public License along with this
program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330,
Boston, MA  02111-1307  USA

*/

#include "qrccrossover.h"
#include "../../encodings/qrcindividual.h"

#include <math.h>

using namespace ECWorkshop;

///////////////////////////////////////////////////////////
//
//		QRCCrossover class
//
///////////////////////////////////////////////////////////

QRCCrossover::QRCCrossover (void) : _alpha(0.0) {}

QRCCrossover::QRCCrossover (const QEvolutionaryAlgorithm* owner)
: QCrossoverOperator(owner), _alpha(0.0) {
}

QRCCrossover::QRCCrossover (const QRCCrossover& argX) : QCrossoverOperator(argX), _alpha(argX._alpha) {
}

QRCCrossover::~QRCCrossover (void) {
	_alpha = 0.0;
}

QRCCrossover& QRCCrossover::operator= (const QRCCrossover& argX) {
	this->QCrossoverOperator::operator = (argX);
	_alpha = argX._alpha;

	return (*this);
}

QEncodingType QRCCrossover::getEncodingType () const {
	return QET_REAL;
}

int QRCCrossover::setParameters (const QEAParameters& argParams) {
	this->QCrossoverOperator::setParameters (argParams);
	_alpha = _params->xAlpha;
	return Q_OK;
}

///////////////////////////////////////////////////////////
//
//		QRCCrossoverArithmetic class
//
// Notes:
// --------------------------
// Arithmetic crossover operator for [real-coded] encoding
//
///////////////////////////////////////////////////////////

int QRCCrossoverArithmetic::cross (const __VECTOR_CONST_INDIVIDUAL_PTR* argParents, __VECTOR_INDIVIDUAL_PTR* argChildren) {
	int i;
//	int genesNumber = _params->chromosomeLength;
	int genesNumber = (*argParents)[0]->getSize();
	double pw1, pw2;	// parent weights
	double cw1, cw2;	// children weights
	double temp;
	const QRCIndividual* p1 = (const QRCIndividual*)((*argParents)[0]), *p2 = (const QRCIndividual*)((*argParents)[1]);
	QRCIndividual *c1 = (QRCIndividual*)((*argChildren)[0]), *c2 = (QRCIndividual*)((*argChildren)[1]);

	for (i=0; i<genesNumber; i++) {
		pw1 = p1->getGeneValue(i);
		pw2 = p2->getGeneValue(i);

		temp = QRandN();
		cw1 = temp * pw1 + (1.0-temp) * pw2;
		cw2 = temp * pw2 + (1.0-temp) * pw1;
		c1->setGeneValue (i, cw1);
		c2->setGeneValue (i, cw2);
	}

	return Q_OK;
}

QRCCrossoverArithmetic::QRCCrossoverArithmetic (void) {
}

QRCCrossoverArithmetic::QRCCrossoverArithmetic (const QEvolutionaryAlgorithm* owner)
: QRCCrossover(owner) {
}

QRCCrossoverArithmetic::QRCCrossoverArithmetic (const QRCCrossoverArithmetic& argX)
: QRCCrossover(argX) {
}

QRCCrossoverArithmetic::~QRCCrossoverArithmetic (void) {
}

QRCCrossoverArithmetic& QRCCrossoverArithmetic::operator= (const QRCCrossoverArithmetic& argX) {
	this->QRCCrossover::operator = (argX);
	return (*this);
}

QCrossoverOperator* QRCCrossoverArithmetic::clone () const {
	return new QRCCrossoverArithmetic ((QRCCrossoverArithmetic&)*this);
}


///////////////////////////////////////////////////////////
//
//		QRCCrossoverSBX class
//
// Notes:
// --------------------------
// SBX crossover operator for [real-coded] encoding
//
///////////////////////////////////////////////////////////

int QRCCrossoverSBX::cross (const __VECTOR_CONST_INDIVIDUAL_PTR* argParents, __VECTOR_INDIVIDUAL_PTR* argChildren) {
	int i;
	int genesNumber = _params->chromosomeLength;
	double bk,		// randomized parameter
		cw1, cw2;	// children weights
	const QRCIndividual* p1 = (const QRCIndividual*)((*argParents)[0]), *p2 = (const QRCIndividual*)((*argParents)[1]);
	QRCIndividual *c1 = (QRCIndividual*)((*argChildren)[0]), *c2 = (QRCIndividual*)((*argChildren)[1]);

	for (i=0; i<genesNumber; i++) {
		bk = QRandN ();
		if (bk <= 0.5) {
			bk = pow (2*bk, 1.0/(_alpha+1.0));
		} else {
			bk = pow (2*(1.0 - bk), 1.0/(_alpha+1.0));
		}
		cw1 = 0.5*((1-bk)*p1->getGeneValue(i) + (1+bk)*p2->getGeneValue(i));
		cw2 = 0.5*((1+bk)*p1->getGeneValue(i) + (1-bk)*p2->getGeneValue(i));
		c1->setGeneValue (i, cw1);
		c2->setGeneValue (i, cw2);
	}

	return Q_OK;
}

QRCCrossoverSBX::QRCCrossoverSBX (void) {
}

QRCCrossoverSBX::QRCCrossoverSBX (const QEvolutionaryAlgorithm* owner)
: QRCCrossover(owner) {
}

QRCCrossoverSBX::QRCCrossoverSBX (const QRCCrossoverSBX& argX)
: QRCCrossover(argX) {
}

QRCCrossoverSBX::~QRCCrossoverSBX (void) {
}

QRCCrossoverSBX& QRCCrossoverSBX::operator= (const QRCCrossoverSBX& argX) {
	this->QRCCrossover::operator = (argX);
	return (*this);
}

QCrossoverOperator* QRCCrossoverSBX::clone () const {
	return new QRCCrossoverSBX ((QRCCrossoverSBX&)*this);
}


///////////////////////////////////////////////////////////
//
//		QRCCrossoverBLX class
//
// Notes:
// --------------------------
// BLX crossover operator for [real-coded] encoding
//
///////////////////////////////////////////////////////////

int QRCCrossoverBLX::cross (const __VECTOR_CONST_INDIVIDUAL_PTR* argParents, __VECTOR_INDIVIDUAL_PTR* argChildren) {
	int i;
//	int genesNumber = _params->chromosomeLength;
	int genesNumber = (*argParents)[0]->getSize();
	double deltaW;
	double pw1, pw2;	// parent weights
	double cw1, cw2;	// children weights
	const QRCIndividual* p1 = (const QRCIndividual*)((*argParents)[0]), *p2 = (const QRCIndividual*)((*argParents)[1]);
	QRCIndividual *c1 = (QRCIndividual*)((*argChildren)[0]), *c2 = (QRCIndividual*)((*argChildren)[1]);

	for (i=0; i<genesNumber; i++) {
		pw1 = p1->getGeneValue(i);
		pw2 = p2->getGeneValue(i);
		if (pw1 > pw2) {
			swap (pw1, pw2);
		}
//		cout<<pw1<<" : "<<pw2<<"\n";
		deltaW = pw2 - pw1;
//		cout<<deltaW<<"\n";
		pw1 -= _alpha*deltaW;
		pw2 += _alpha*deltaW;
		deltaW = (pw2 - pw1);
//		cout<<deltaW<<"\n";

		cw1 = QRandN() * deltaW + pw1;	// pw1 is used as a lower bound for interval
		cw2 = QRandN() * deltaW + pw1;
		c1->setGeneValue (i, cw1);
		c2->setGeneValue (i, cw2);
	}

	return Q_OK;
}

QRCCrossoverBLX::QRCCrossoverBLX (void) {}

QRCCrossoverBLX::QRCCrossoverBLX (const QEvolutionaryAlgorithm* owner)
: QRCCrossover(owner) {
}

QRCCrossoverBLX::QRCCrossoverBLX (const QRCCrossoverBLX& argX)
: QRCCrossover(argX) {
}

QRCCrossoverBLX::~QRCCrossoverBLX (void) {}

QRCCrossoverBLX& QRCCrossoverBLX::operator= (const QRCCrossoverBLX& argX) {
	this->QRCCrossover::operator= (argX);

	return (*this);
}

QCrossoverOperator* QRCCrossoverBLX::clone () const {
	return new QRCCrossoverBLX ((QRCCrossoverBLX&)*this);
}
