/*****************************************

ECWorkshop -- evolutionary computation
toolkit library

Web: http://qai.narod.ru/ecw/
Mailto: qai@mail.ru

Copyright (C) 2007 by Yury Tsoy


This program is free software; you can
redistribute it and/or modify it under the
terms of the GNU General Public License as
published by the Free Software Foundation;
either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that
it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU General Public License
for more details.

You should have received a copy of the GNU
General Public License along with this
program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330,
Boston, MA  02111-1307  USA

*/

#include "qannsin.h"
#include "../encodings/include.h"
#include "../operators/include.h"

#include <math.h>

using namespace ECWorkshop;

QANNProblemSinDouble::QANNProblemSinDouble (void) {
	// set number of inputs and outputs
	_inputsNumber = 1;
	_outputsNumber = 1;

	// fill training vectors
	_inputs.resize (400);
	_outputs.resize (400);
	_inputs[0].push_back (1);
	_outputs[0].push_back (sin(0.2) + sin(0.311));
	for (int i=1; i<400; i++) {
		_inputs[i].push_back (_outputs[i-1][0]);
		_outputs[i].push_back (sin(0.2*(i+1)) + sin(0.311*(i+1)));
	}

	// fill test data
	_testI.resize (300);
	_testO.resize (300);
	_testI[0] = _outputs[399][0];
	_testO[0] = sin(0.2*401.0) + sin(0.311*401.0);
	for (int i=1; i<300; i++) {
		_testI[i] = _testO[i-1];
		_testO[i] = sin(0.2*(i+401)) + sin(0.311*(i+401));
	}
}

QANNProblemSinDouble::~QANNProblemSinDouble (void) {
	_inputs.clear ();
	_outputs.clear ();
}

// argN -- index of the training sample
double QANNProblemSinDouble::calculateError (__VECTOR_DOUBLE& argV, const int argN) const {
	double er = 0.0, temp;

	// enlarge [argV] if necessary
	//while ((int)argV.size() < _outputsNumber) {
	//	argV.push_back (0);
	//}
	temp = argV[_inputsNumber+1] - _outputs[argN][0];
	er += temp*temp;

	return er;
}


// argN -- index of the training sample
double QANNProblemSinDouble::calculateErrorTest (__VECTOR_DOUBLE& argV, const int argN) const {
	double er = 0.0, temp;

	// enlarge [argV] if necessary
	//while ((int)argV.size() < _outputsNumber) {
	//	argV.push_back (0);
	//}
	temp = argV[_inputsNumber+1] - _testO[argN];
	er += temp*temp;

	return er;
}


// int evaluateIndividual (QIndividual* argInd)
// Evaluates given individual
//	Returns Q_OK on success
int QANNProblemSinDouble::evaluateIndividual (QIndividual* argInd) {
	if (argInd->getEncodingType() == QET_REGULATORY_NETWORK) {
		return evaluateIndividualRN (argInd);
	}
	return Q_FAILURE;
}

int QANNProblemSinDouble::evaluateIndividualRN (QIndividual* argInd) {
	QRNIndividual* tempInd = (QRNIndividual*)argInd;
	__VECTOR_DOUBLE output, input;
	QFitness fitness;
	int size = (int)_inputs.size ();
	int i, minD, maxD;
	double nodesN = tempInd->getSize();
	double er = 0.0, energy = 0.0;

	tempInd->calculateMinMaxDistances (minD, maxD);

	if (minD > 0) {
		// washout the first 100 steps
		for (i=0; i<100; ++i) {
			if (i > 0) {
//				output[2] = _inputs[i][0]; // use teacher forcing
				input.assign (output.begin(), output.end ());
				output.clear ();
				input[1] = _inputs[i][0];	// use teacher forcing
				//output[2] = 0;	// reset output
//				input[1] = _inputs[i][0] + 0.01*QRandGauss();	// use noisy teacher forcing
				//input[1] = output[2];
				//input[1] = 0;	// toggle signal 1 off
			} else {
				input.clear ();
				//input.assign (2, 0);
				input.push_back (0);
				input.push_back (_inputs[0][0]);
			}

			tempInd->calculate1step (input, output);
			//output.clear ();
			//tempInd->calculateNsteps (input, output, maxD);
		}
		for (i=100; i<size; i++) {
			//output[2] = _inputs[i][0]; // use teacher forcing
			input.assign (output.begin(), output.end ());
			//input[1] = output[2];
			output.clear ();	// reset the new state
			input[1] = _inputs[i][0];	// set input signal using teacher's choice
			//input[1] = _inputs[i][0] + 0.01*QRandGauss();	// use noisy teacher forcing

			tempInd->calculate1step (input, output);
			//output.clear ();
			//tempInd->calculateNsteps (input, output, maxD);

			if (output.empty ()) {
				output.assign (_inputsNumber + _outputsNumber + 1, 0);
			}

			energy += tempInd->getEnergy ();

			// calculate output error
			er += calculateError (output, i);
		}
		energy /= (double)300.0;
		er /= (double)300.0;
//		er *= 100.0;
	} else {
		// The input is disconnected with the output. No calculation is performed
		er = 100.0;	//0.25;
		energy = 100.0;
	}	// if (minD > 0)

	energy = 0.01*fabs(nodesN - energy)/nodesN;
//	energy *= 0.001;
	nodesN = log (nodesN) * 0.001 + 0.001;

	if (er < _params->errorLevel) {
		fitness.setValue (er);
	} else {
		fitness.setValue (er + nodesN);
	}
	fitness.addExtra (er);
	fitness.addExtra (energy);
	fitness.addExtra (nodesN);
	tempInd->setFitness (fitness);

	return Q_OK;
}

// int evaluatePopulation (QPopulation* argPopul)
// Evaluates all individuals in given population
//	Returns Q_OK on success
int QANNProblemSinDouble::evaluatePopulation (QPopulation* argPopul) {
	if (argPopul->getEncodingType () == QET_REGULATORY_NETWORK) {
		int i, popSize = argPopul->getSize ();

		for (i=0; i<popSize; i++) {
			evaluateIndividual (argPopul->getIndividual( i ));
		}

		return Q_OK;
	} else {
	        // shout aloud and ...
		cerr<<"\n\n[QANNProblemXOR::evaluatePopulation] error:\n"
			<<"\tIncorrect encoding type! (encoding ID = "<<argPopul->getEncodingType()<<")\n\n";
	        // ... do nothing
		return Q_FAILURE;
	}
}

// int printProblemName (ostream& out)
// Prints name of the problem
//	Returns Q_OK on success
int QANNProblemSinDouble::printProblemName (ostream& out) {
	out<<"Neural network double sin prediction problem\n";
	return Q_OK;
}

// Sets [argParams] to default values according to the problem at hand
// In most cases just ignore this function unless you need to set
//  some EA's parameters that depend on the problem
int QANNProblemSinDouble::getDefaultParameters (QEAParameters* argParams) {
	argParams->inputs = 1;
	argParams->outputs = 1;
	//argParams->encoding = QET_REGULATORY_NETWORK;
	return Q_OK;
}

int QANNProblemSinDouble::testIndividual (QIndividual* argInd) const {
	if (argInd->getEncodingType() == QET_REGULATORY_NETWORK) {
		QRNIndividual* tempInd = (QRNIndividual*)argInd;
		__VECTOR_DOUBLE input, output;
		int size = (int)_inputs.size();
		int i, minD, maxD;
		double er = 0.0, energy = 0.0, temp;

		ofstream fout;
		fout.open ("qannsin.log");
		fout<<"Network output, testing:\n";

		tempInd->calculateMinMaxDistances (minD, maxD);

		if (minD > 0) {

			// introduce training sets with teacher forcing
			size = (int)_inputs.size ();
			//for (i=0; i<size; i++) {
			//	if (i > 0) {
			//		// update current state
			//		input.assign (output.begin (), output.end());
			//		// substitute input signal by teacher's choice
			//		input[1] = _inputs[i][0];
			//		//input[1] = _inputs[i][0] + 0.01*QRandGauss();	// use noisy teacher forcing
			//	} else {
			//		// init environment
			//		input.clear ();
			//		input.push_back (0);
			//		input.push_back (1);
			//	}

			//	tempInd->calculate1step (input, output);
			//	//output.clear ();
			//	//tempInd->calculateNsteps (input, output, maxD);

			//	if (output.empty ()) {
			//		output.assign (_inputsNumber + _outputsNumber + 1, 0);
			//	}

			//	energy += tempInd->getEnergy ();

			//	// calculate output error
			//	er += calculateError (output, i);
			//	fout<<input[1]<<"\t"<<output[_inputsNumber+1]<<"\t"<<_outputs[i][0]<<"\t"<<calculateError (output, i)<<"\n";
			//}	// for (i=0; i<size; i++)
			//er /= (double)size;

			//***************************************************************
			//
			// introduce training sets
			fout<<"Testing:\n";
			er = 0.0;
			// reset environment state
			input.clear ();
			output.clear ();
			// ignore washout time
			for (i=0; i<100; ++i) {
				if (i > 0) {
					//output[2] = _inputs[i][0]; // use teacher forcing
					input.assign (output.begin (), output.end());
					output.clear ();
					input[1] = _inputs[i][0];	// copy output to input
					//output[2] = 0;	// reset output
					//input[1] = 0;	// toggle signal 1 off
				} else {
					input.clear ();
					input.push_back (0);
					input.push_back (_inputs[0][0]);
//					input.assign (2, 0);
				}
				tempInd->calculate1step (input, output);
				//output.clear ();
				//tempInd->calculateNsteps (input, output, maxD);

				for (int j=2; j<(int)output.size(); j++) {
					fout<<output[j]<<"\t";
				}
				fout<<"\n";
				//fout<<input[1]<<"\t"<<output[_inputsNumber+1]<<"\t"<<_outputs[i][0]<<"\t"<<calculateError (output, i)<<"\n";
			}
			for (i=100; i<size; i++) {
				//output[2] = _inputs[i][0]; // use teacher forcing
				input.assign (output.begin (), output.end());
				//input[1] = output[2];
				output.clear ();
				input[1] = _inputs[i][0];	// copy output to input

				tempInd->calculate1step (input, output);
				//output.clear ();
				//tempInd->calculateNsteps (input, output, maxD);

				if (output.empty ()) {
					output.assign (_inputsNumber + _outputsNumber + 1, 0);
				}

				energy += tempInd->getEnergy ();

				// calculate output error
				er += calculateError (output, i);

				for (int j=2; j<(int)output.size(); j++) {
					fout<<output[j]<<"\t";
				}
				fout<<"\n";
				//fout<<input[1]<<"\t"<<output[_inputsNumber+1]<<"\t"<<_outputs[i][0]<<"\t"<<calculateError (output, i)<<"\n";
			}

			// prediction stage
			size = (int)_testI.size ();
			for (i=0; i<size; i++) {
				input.assign (output.begin (), output.end());
				input[1] = output [2];	// copy output to input
				output.clear ();
				//output[2] = 0;	// reset output
				//input[1] = _testO[i];	// use teacher forcing
				//input[1] = 0;	// toggle signal 1 off

				tempInd->calculate1step (input, output);
				//output.clear ();
				//tempInd->calculateNsteps (input, output, maxD);

				if (output.empty ()) {
					output.assign (_inputsNumber + _outputsNumber + 1, 0);
				}

				energy += tempInd->getEnergy ();

				// calculate output error
				er += calculateErrorTest (output, i);

				for (int j=2; j<(int)output.size(); j++) {
					fout<<output[j]<<"\t";
				}
				fout<<"\n";
				//fout<<input[1]<<"\t"<<output[_inputsNumber+1]<<"\t"<<_testO[i]<<"\t"<<calculateErrorTest (output, i)<<"\n";
			}

			//// introduce test sets
			//size = (int)_testI.size ();
			//for (i=0; i<size; i++) {
			//	input.assign (output.begin(), output.end());
			//	input[1] = output[2];

			//	//for (int j=0; j<input.size(); ++j) {
			//	//	cout<<" "<<input[j];
			//	//}
			//	//cout<<"\n";

			//	tempInd->calculate1step (input, output);
			//	//output.clear ();	// to terminate the influence from the previous time-steps
			//	//tempInd->calculateNsteps (input, output, maxD);

			//	if (output.empty ()) {
			//		output.assign (_inputsNumber + _outputsNumber + 1, 0.5);
			//	}

			//	// calculate energy
			//	energy += tempInd->getEnergy ();

			//	// calculate output error
			//	temp = calculateErrorTest (output, i);
			//	cout<<"Test input set # "<<(i+1)<<". Output = "<<output[2]<<". Error = "<<temp<<"\n";
			//	er += temp;

			//	fout<<input[1]<<"\t"<<output[2]<<"\t"<<_testO[i]<<"\t"<<calculateErrorTest (output, i)<<"\n";
			//}
			er /= (double)(size*2.0);
			cout<<"Mean error = "<<er<<"\n";
			energy /= (double)size;
			cout<<"Mean energy = "<<energy<<"\n";
		} else {
			er = 0;
			cout<<"Incorrect network structure: The output is unreachable\n";
		}

		fout<<"\nError:\t"<<er<<"\n";

		return Q_OK;
	} else {
		cerr<<"\n\n[QANNProblemSinDouble::testIndividual] error:\n"
			<<"\tIncorrect encoding type! (encoding ID = "<<argInd->getEncodingType()<<")\n\n";
	        // ... do nothing
		return Q_FAILURE;
	}
}

