/*****************************************

ECWorkshop -- evolutionary computation
toolkit library

Web: http://qai.narod.ru/ecw/
Mailto: qai@mail.ru

Copyright (C) 2007 by Yury Tsoy


This program is free software; you can
redistribute it and/or modify it under the
terms of the GNU General Public License as
published by the Free Software Foundation;
either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that
it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU General Public License
for more details.

You should have received a copy of the GNU
General Public License along with this
program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330,
Boston, MA  02111-1307  USA

*/


#include "qannxor.h"
#include "../encodings/include.h"
#include "../operators/include.h"

#include <math.h>

using namespace ECWorkshop;

QANNProblemXOR::QANNProblemXOR (void) {
	// set number of inputs and outputs
	_inputsNumber = 2;
	_outputsNumber = 1;

	// set training data of 4 vectors sets
	_inputs.resize (4);
	_outputs.resize (4);

	// fill training vectors
	_inputs[0].resize (_inputsNumber);
	_outputs[0].resize (_outputsNumber);
	_inputs[0][0] = 0;
	_inputs[0][1] = 0;
	_outputs[0][0] = 0;

	_inputs[1].resize (_inputsNumber);
	_outputs[1].resize (_outputsNumber);
	_inputs[1][0] = 0;
	_inputs[1][1] = 1;
	_outputs[1][0] = 1;

	_inputs[2].resize (_inputsNumber);
	_outputs[2].resize (_outputsNumber);
	_inputs[2][0] = 1;
	_inputs[2][1] = 0;
	_outputs[2][0] = 1;

	_inputs[3].resize (_inputsNumber);
	_outputs[3].resize (_outputsNumber);
	_inputs[3][0] = 1;
	_inputs[3][1] = 1;
	_outputs[3][0] = 0;
}

QANNProblemXOR::~QANNProblemXOR (void) {
	_inputs.clear ();
	_outputs.clear ();
}

// argN -- index of the training sample
double QANNProblemXOR::calculateError (__VECTOR_DOUBLE& argV, const int argN) const {
	double er = 0.0, temp;
	int i;

	// enlarge [argV] if necessary
	while ((int)argV.size() < _outputsNumber) {
		argV.push_back (0);
	}
	for (i=0; i<_outputsNumber; i++) {
		//cout<<argN<<":\t"<<i<<" : "<<argV[i]<<"\n";
		temp = argV[i] - _outputs[argN][i];
		er += temp*temp;
	}

	return er;
}


// int evaluateIndividual (QIndividual* argInd)
// Evaluates given individual
//	Returns Q_OK on success
int QANNProblemXOR::evaluateIndividual (QIndividual* argInd) {
	if (argInd->getEncodingType() == QET_REGULATORY_NETWORK) {
		return evaluateIndividualRN (argInd);
	}
	return Q_FAILURE;
}

int QANNProblemXOR::evaluateIndividualRN (QIndividual* argInd) {
	QRNIndividual* tempInd = (QRNIndividual*)argInd;
	__VECTOR_DOUBLE output, input;
	QFitness fitness;
	int size = (int)_inputs.size ();
	int i, minD, maxD;
	double nodesN = tempInd->getSize();
	double er = 0.0, energy = 0.0;

	tempInd->calculateMinMaxDistances (minD, maxD);

	if (minD > 0) {
		for (i=0; i<size; i++) {
			input.assign (_inputs[i].begin (), _inputs[i].end());
			input.insert (input.begin(), 0);

			//for (int j=0; j<input.size(); ++j) {
			//	cout<<" "<<input[j];
			//}
			//cout<<"\n";

			output.clear ();
			tempInd->calculateNsteps (input, output, maxD);

			if (output.empty ()) {
				output.assign (_outputsNumber, 0);
			} else {
				// delete information on input signals
				if ((int)output.size() <= (_inputsNumber + 1)) {
					tempInd->print2Stream(cout);
					cout<<"Calculation steps number:"<<maxD<<"\n";
					cout<<"Minimal distance:"<<minD<<"\n";
					cout<<"Strange output vector:\n";
					for (int j=0; j<(int)output.size(); ++j) {
						cout<<" "<<output[j];
					}
					cout<<"\n";
					exit (0);
				}
				output.erase (output.begin(), output.begin() + _inputsNumber + 1);
			}

			// convert result into appropriate form
			for (int j=0; j<(int)output.size(); ++j) {
				//cout<<" "<<output[j];
				output[j] = 0.5*(output[j] + 1.0);
				if (output[j] < 0) {
					output[j] = 0;
				} else if (output[j] > 1) {
					output[j] = 1;
				}
				//cout<<" -> "<<output[j];
			}
			//cout<<"\n";

			energy += tempInd->getEnergy ();

			// calculate output error
			er += calculateError (output, i);
		}
		energy /= (double)(size);
		er /= (double)size;
//		er *= 100.0;
	} else {
		// The input is disconnected with the output. No calculation is performed
		er = 1.0;	//0.25;
		energy = 100.0;
	}	// if (minD > 0)

	energy = 0.01*fabs(nodesN - energy)/nodesN;
//	energy *= 0.001;
	nodesN = log (nodesN) * 0.001 + 0.001;

	if (er < _params->errorLevel) {
		fitness.setValue (er);
	} else {
		fitness.setValue (er + nodesN);
	}
	fitness.addExtra (er);
	fitness.addExtra (energy);
	fitness.addExtra (nodesN);
	tempInd->setFitness (fitness);

	return Q_OK;
}

// int evaluatePopulation (QPopulation* argPopul)
// Evaluates all individuals in given population
//	Returns Q_OK on success
int QANNProblemXOR::evaluatePopulation (QPopulation* argPopul) {
	if (argPopul->getEncodingType () == QET_REGULATORY_NETWORK) {
		int i, popSize = argPopul->getSize ();

		for (i=0; i<popSize; i++) {
			evaluateIndividual (argPopul->getIndividual( i ));
		}

		return Q_OK;
	} else {
	        // shout aloud and ...
		cerr<<"\n\n[QANNProblemXOR::evaluatePopulation] error:\n"
			<<"\tIncorrect encoding type! (encoding ID = "<<argPopul->getEncodingType()<<")\n\n";
	        // ... do nothing
		return Q_FAILURE;
	}
}

// int printProblemName (ostream& out)
// Prints name of the problem
//	Returns Q_OK on success
int QANNProblemXOR::printProblemName (ostream& out) {
	out<<"Neural network XOR problem\n";
	return Q_OK;
}

// Sets [argParams] to default values according to the problem at hand
// In most cases just ignore this function unless you need to set
//  some EA's parameters that depend on the problem
int QANNProblemXOR::getDefaultParameters (QEAParameters* argParams) {
	argParams->inputs = 2;
	argParams->outputs = 1;
	argParams->encoding = QET_REGULATORY_NETWORK;
	return Q_OK;
}

int QANNProblemXOR::testIndividual (QIndividual* argInd) const {
	if (argInd->getEncodingType() == QET_REGULATORY_NETWORK) {
		QRNIndividual* tempInd = (QRNIndividual*)argInd;
		__VECTOR_DOUBLE input, output;
		int size = (int)_inputs.size();
		int i, minD, maxD;
		double er = 0.0, energy = 0.0, temp;

		tempInd->calculateMinMaxDistances (minD, maxD);

		if (minD > 0) {

			for (i=0; i<size; i++) {
				input.assign (_inputs[i].begin (), _inputs[i].end());
				input.insert (input.begin(), 0);

				//for (int j=0; j<input.size(); ++j) {
				//	cout<<" "<<input[j];
				//}
				//cout<<"\n";

				output.clear ();	// to terminate the influence from the previous time-steps
				tempInd->calculateNsteps (input, output, maxD);

				if (output.empty ()) {
					output.assign (_outputsNumber, 0.5);
				} else {
					// delete information on input signals
					if ((int)output.size() <= (_inputsNumber + 1)) {
						tempInd->print2Stream(cout);
						cout<<"Calculation steps number:"<<maxD<<"\n";
						cout<<"Minimal distance:"<<minD<<"\n";
						cout<<"Strange output vector:\n";
						for (int j=0; j<(int)output.size(); ++j) {
							cout<<" "<<output[j];
						}
						cout<<"\n";
						exit (0);
					}
					output.erase (output.begin(), output.begin() + _inputsNumber + 1);
				}

				// convert result into appropriate form
				for (int j=0; j<(int)output.size(); ++j) {
					//cout<<" "<<output[j];
					output[j] = 0.5*(output[j] + 1.0);
					if (output[j] < 0) {
						output[j] = 0;
					} else if (output[j] > 1) {
						output[j] = 1;
					}
					//cout<<" -> "<<output[j];
				}
				//cout<<"\n";

				// calculate energy
				energy += tempInd->getEnergy ();

				// calculate output error
				temp = calculateError (output, i);
				cout<<"Test input set # "<<(i+1)<<". Output = "<<output[0]<<". Error = "<<temp<<"\n";
				er += temp;
			}
			er /= (double)size;
			cout<<"Mean error = "<<er<<"\n";
			energy /= (double)size;
			cout<<"Mean energy = "<<energy<<"\n";
		} else {
			er = 0;
			cout<<"Incorrect network structure: The output is unreachable\n";
		}

		return Q_OK;
	} else {
		cerr<<"\n\n[QANNProblemXOR::testIndividual] error:\n"
			<<"\tIncorrect encoding type! (encoding ID = "<<argInd->getEncodingType()<<")\n\n";
	        // ... do nothing
		return Q_FAILURE;
	}
}
