/*****************************************

ECWorkshop -- evolutionary computation
toolkit library

Web: http://qai.narod.ru/ecw/
Mailto: qai@mail.ru

Copyright (C) 2007 by Yury Tsoy


This program is free software; you can
redistribute it and/or modify it under the
terms of the GNU General Public License as
published by the Free Software Foundation;
either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that
it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU General Public License
for more details.

You should have received a copy of the GNU
General Public License along with this
program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330,
Boston, MA  02111-1307  USA

*/

#include "qsummaryprocessor.h"

#include "../eas/include.h"
#include "../encodings/include.h"

#include <algorithm>
#include <fstream>

using namespace ECWorkshop;

QSummaryProcessor::QSummaryProcessor (void) {
}

QSummaryProcessor::~QSummaryProcessor(void) {
	__VECTOR_RUN_SUMMARY::iterator iter = _summaries.begin(), end = _summaries.end();
	for (; iter != end; iter++) {
		(*iter).reset ();
	}
	_summaries.clear ();
}

int QSummaryProcessor::setParameters (const QEAParameters* argParams) {
	_params = argParams;
	return Q_OK;
}

int QSummaryProcessor::addSummary (const QRunSummary& argSummary) {
	_summaries.push_back (argSummary);
	return (int)_summaries.size();
}

int QSummaryProcessor::getSummariesNumber () {
	return (int)_summaries.size ();
}

int QSummaryProcessor::operate () {
	if (_summaries.empty() == false) {
		ofstream fout;
		QRunSummary result;
		int i, count;
		int sumNumber = (int)_summaries.size();
		double mean, dev;
		__VECTOR_DOUBLE devMean, devDev, devBest, devWorst;	// deviation values for means and deviations at each generation calculated over all runs
		__VECTOR_DOUBLE tempV;
		__VECTOR_RUN_SUMMARY::iterator iter = _summaries.begin ();

		fout.open ("summary.log");

		// ************
		// calculate average fitness values over all runs
		result.means.resize (_params->generationsNumber);
		result.devs.resize (_params->generationsNumber);
		result.bests.resize (_params->generationsNumber);
		result.worsts.resize (_params->generationsNumber);
		result.popSizes.resize (_params->generationsNumber);
		devMean.assign (_params->generationsNumber, 0.0);
		devDev.assign (_params->generationsNumber, 0.0);
		devBest.assign (_params->generationsNumber, 0.0);
		devWorst.assign (_params->generationsNumber, 0.0);
		for (i=0; i<_params->generationsNumber; i++) {
			result.means[i] = 0;
			result.devs[i] = 0;
			result.bests[i] = 0;
			result.worsts[i] = 0;
			result.popSizes[i] = 0;
			count = 0;

			for (iter = _summaries.begin(); iter != _summaries.end(); iter++) {
				if ((*iter).generation > i) {
					result.means[i] += (*iter).means[i];
					result.devs[i] += (*iter).devs[i];
					result.bests[i] += (*iter).bests[i];
					result.worsts[i] += (*iter).worsts[i];
					result.popSizes[i] += (*iter).popSizes[i];

					devMean[i] += (*iter).means[i]*(*iter).means[i];
					devDev[i] += (*iter).devs[i]*(*iter).devs[i];
					devBest[i] += (*iter).bests[i]*(*iter).bests[i];
					devWorst[i] += (*iter).worsts[i]*(*iter).worsts[i];

					count++;
				}
			}

			if (count > 0) {
				result.means[i] /= (double)count;
				result.devs[i] /= (double)count;
				result.bests[i] /= (double)count;
				result.worsts[i] /= (double)count;
				result.popSizes[i] /= (double)count;

				devMean[i] = devMean[i]/(double)count - result.means[i]*result.means[i];
				devDev[i] = devDev[i]/(double)count - result.devs[i]*result.devs[i];
				devBest[i] = devBest[i]/(double)count - result.bests[i]*result.bests[i];
				devWorst[i] = devWorst[i]/(double)count - result.worsts[i]*result.worsts[i];
			}
		}

		_params->print2stream (fout);
		fout<<"\n#\tMeans\tDeviations\tBests\tWorsts\tDevMean\tDevDev\tDevBests\tDevWorsts:\n"
			<<"----------------------------------------------------\n";
		for (i=0; i<(int)result.means.size(); i++) {
			fout<<(i+1)<<":\t";
			fout<<result.means[i]<<"\t"<<result.devs[i]<<"\t"<<result.bests[i]<<"\t"<<result.worsts[i]<<"\t"<<devMean[i]<<"\t"<<devDev[i]<<"\t"<<devBest[i]<<"\t"<<devWorst[i]<<"\n";
		}
		fout<<"Min averaged best:\t"<<(*min_element (result.bests.begin(), result.bests.end()))<<"\n";

		// ************
		// calculate average fitness values over last 100 generations
		if ((int)result.means.size() > 100) {
			double mean100, dev100, best100, worst100, devMean100, devDev100, devBest100, devWorst100;
			int size = (int)result.means.size();
			mean100 = dev100 = best100 = worst100 = devMean100 = devDev100 = devBest100 = devWorst100 = 0.0;
			for (i=size-100; i<size; i++) {
				mean100 += result.means[i];
				dev100 += result.devs[i];
				best100 += result.bests[i];
				worst100 += result.worsts[i];
				devMean100 += devMean[i];
				devDev100 += devDev[i];
				devBest100 += devBest[i];
				devWorst100 += devWorst[i];
			}
			mean100 /= 100.0;
			dev100 /= 100.0;
			best100 /= 100.0;
			worst100 /= 100.0;
			devMean100 /= 100.0;
			devDev100 /= 100.0;
			devBest100 /= 100.0;
			devWorst100 /= 100.0;
			fout<<"\nAverage values over last 100 generations:\n"
				<<"Mean\tDeviation\tBest\tWorst\tdevMean\tdevDev\tdevBest\tdevWorst:\n"
				<<"----------------------------------------------------\n"
				<<mean100<<"\t"<<dev100<<"\t"<<best100<<"\t"<<worst100<<"\t"<<devMean100<<"\t"<<devDev100<<"\t"<<devBest100<<"\t"<<devWorst100<<"\n";
		}

		// ************
		// print population size dynamics
		mean = dev = 0.0;
		fout<<"\n#\tPopulation size:\n"
			<<"----------------------------------------------------\n";
		for (i=0; i<(int)result.popSizes.size(); i++) {
			fout<<(i+1)<<":\t"<<result.popSizes[i]<<"\n";
		}
		qstats (result.popSizes, mean, dev);
		fout<<"\nMean:\t"<<mean<<"\n"
			<<"Deviation:\t"<<dev<<"\n";

		// ************
		// process times per run statistics
		tempV.clear ();
		for (iter = _summaries.begin(); iter != _summaries.end(); iter++) {
			tempV.push_back ((*iter).time);
		}
		qstats (tempV, mean, dev);

		fout<<"\n#\tTimes:\n"
			<<"------------------\n";
		for (i=0; i<(int)tempV.size(); i++) {
			fout<<(i+1)<<":\t"<<tempV[i]<<"\n";
		}
		fout<<"\nMean:\t"<<mean<<"\n"
			<<"Deviation:\t"<<dev<<"\n";

		// process convergence times per run statistics
		tempV.clear ();
		count = 0;
		mean = dev = 0.0;
		for (iter = _summaries.begin(); iter != _summaries.end(); iter++) {
			if ((*iter).convergenceGeneration != 0) {
				tempV.push_back ((*iter).convergenceGeneration);
				mean += (*iter).convergenceGeneration;
				dev += (*iter).convergenceGeneration * (*iter).convergenceGeneration;
				count++;
			}
		}
		if (count > 0) {	// write statistics only if there is something to write
			mean /= (double)count;
			dev = dev/(double)count - mean*mean;

			fout<<"\n#\tConvergence times:\n"
				<<"------------------\n";
			for (i=0; i<(int)tempV.size(); i++) {
				fout<<(i+1)<<":\t"<<tempV[i]<<"\n";
			}
			fout<<"\nMean:\t"<<mean<<"\n"
				<<"Deviation:\t"<<dev<<"\n";
		} else {
			fout<<"\nNo convergence\n";
		}

		// process solution search statistics
		double meanTime, meanGen, meanFE, devTime, devGen, devFE;
		tempV.clear ();
		count = 0;
		meanTime = meanGen = meanFE = devTime = devGen = devFE = 0.0;
		for (iter = _summaries.begin(); iter != _summaries.end(); iter++) {
			if ((*iter).firstHitFECount != 0) {
//				tempV.push_back ((*iter).firstHitTime);
				meanTime += (*iter).firstHitTime;
				devTime += (*iter).firstHitTime * (*iter).firstHitTime;
				meanGen += (*iter).firstHitGeneration;
				devGen += (*iter).firstHitGeneration * (*iter).firstHitGeneration;
				meanFE += (*iter).firstHitFECount;
				devFE += (*iter).firstHitFECount * (*iter).firstHitFECount;
				count++;
			}
		}
		if (count > 0) {	// write statistics only if there is something to write
			meanTime /= (double)count;
			devTime = devTime/(double)count - meanTime*meanTime;
			meanGen /= (double)count;
			devGen = devGen/(double)count - meanGen*meanGen;
			meanFE /= (double)count;
			devFE = devFE/(double)count - meanFE*meanFE;

			fout<<"\nTimes to find a solution:\n"
				<<"\n#\tTime\tGenerations\tFE\n"
				<<"----------------------------\n";
			for (i=0; i<(int)_summaries.size(); i++) {
				fout<<(i+1)<<":\t"<<_summaries[i].firstHitTime<<"\t"<<_summaries[i].firstHitGeneration<<"\t"<<_summaries[i].firstHitFECount<<"\n";
			}
			fout<<"\nMean time:\t"<<meanTime<<"\n"
				<<"Time deviation:\t"<<devTime<<"\n"
				<<"\nMean generation:\t"<<meanGen<<"\n"
				<<"Generation deviation:\t"<<devGen<<"\n"
				<<"\nMean FE count:\t"<<meanFE<<"\n"
				<<"FE count deviation:\t"<<devFE<<"\n";
		} else {
			fout<<"\nNo solution(s) found\n";
		}
		// print algorithm specific results
		fout<<"\n";	// make a gap
		_params->ea->printResults (_summaries, fout);
		fout.close ();

		// print solutions (or candidate-solutions) if there are any
		if (count > 0) {
			fout.open ("solution.log");
			_params->print2stream (fout);
			fout<<"\nSolutions found:\t"<<count<<"\n";
			for (i=0; i<(int)_summaries.size(); i++) {
				if (_summaries[i].firstHitTime != 0) {
					fout<<"\n> Run #"<<(i+1)<<" solution:\n";
					_summaries[i].firstHitSolution->print2Stream (fout);
				}
			}
			fout.close ();
		}
		// print candidate-solutions if there are any
		if (count < (int)_summaries.size()) {
			fout.open ("candidate.log");
			_params->print2stream (fout);
			fout<<"\nCandidate-solutions found:\t"<<((int)_summaries.size()-count)<<"\n";
			for (i=0; i<(int)_summaries.size(); i++) {
				if (_summaries[i].firstHitTime == 0) {
					fout<<"\n> Run #"<<(i+1)<<" candidate-solution:\n";
					_summaries[i].currentBest->print2Stream (fout);
				}
			}
			fout.close ();
		}

		return Q_OK;
	} else {
		cerr<<"\n\n[QSummaryProcessor::operate] error:\n"
			<<"\tNo data to process! (summaries number = "<<(int)_summaries.size()<<")\n\n";
		return Q_FAILURE;
	}	// if (_summaries.empty() == false)
}


int QSummaryProcessor::printParameters (ostream& out) {
	out<<"EA parameters:\n"
		<<"------------------------------\n";

	// encoding parameters
	out<<"Encoding type:\t";
	switch (_params->encoding) {
		case QET_INTEGER:
			out<<"[integer]\n";
			out<<"Gene size (bits):\t"<<_params->geneSize<<"\n";
			out<<"Gene offset:\t"<<_params->geneOffset<<"\n";
			out<<"Gene precision:\t"<<_params->genePrecision<<"\n";
			break;
		case QET_STRING:
			out<<"[string]\n";
			out<<"Arity:\t"<<_params->arity<<"\n";
			break;
		default:
			out<<"[unknown]\n";
			break;
	}
	out<<"Genes number:\t"<<_params->chromosomeLength<<"\n";

	// operators parameters
	out<<"Selection operator:\t";
	switch (_params->selectionType) {
		case QST_NEUTRAL:
			out<<"neutral\n";
			break;
		case QST_TOURNAMENT:
			out<<"tournament\n";
			out<<"Tournament size:\t"<<_params->tournamentSize<<"\n";
			break;
		default:
			out<<"[unknown]\n";
			break;
	}

	out<<"Crossover operator:\t";
	switch (_params->xType) {
		case QXT_1POINT:
			out<<"1-point\n";
			break;
		case QXT_2POINT:
			out<<"2-point\n";
			break;
		case QXT_UNIFORM:
			out<<"uniform\n";
			break;
		default:
			out<<"[unknown]\n";
			break;
	}
	out<<"Crossover probability:\t"<<_params->xRate<<"\n";

	out<<"Mutation operator:\t";
	switch (_params->mutationType) {
		case QMT_BITFLIP:
			out<<"bit-flip\n";
			break;
		case QMT_RANDOMFLIP:
			out<<"random-flip\n";
			break;
		default:
			out<<"[unknown]\n";
			break;
	}
	out<<"Mutation probability:\t"<<_params->mRate<<"\n";

	out<<"Population sizing operator:\t";
	switch (_params->populationSizerType) {
		case QPSZT_CONST:
			out<<"Constant\n";
			break;
		case QPSZT_STEADY:
			out<<"Steady\n";
			out<<"Population sizing step:\t"<<_params->sizingStep<<"\n"
				<<"Population sizing direction:\t"<<_params->sizingDirection<<"\n";
			break;
		case QPSZT_SIMPLE:
			out<<"Simple adaptive\n";
			break;
		default:
			out<<"[unknown]\n";
			break;
	}
	out<<"\n";

	return Q_OK;
}
